"""
`cirrina` - Opinionated web framework

Package file

:license: LGPL, see LICENSE for details
"""

from .server import Server

# expose server class
__all__ = ['Server']

# define package metadata
__VERSION__ = '0.1.0'
